// Copyright (C) 2020-2023 Joel Rosdahl and other contributors
//
// See doc/AUTHORS.adoc for a complete list of contributors.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 51
// Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

#pragma once

#include <util/Fd.hpp>

#include <third_party/tl/expected.hpp>

#include <filesystem>
#include <string>
#include <string_view>

namespace util {

// This class represents a unique temporary file created by mkstemp. The file is
// not deleted by the destructor.
class TemporaryFile
{
public:
  static constexpr char tmp_file_infix[] = ".tmp.";

  TemporaryFile() = delete;

  TemporaryFile(TemporaryFile&& other) noexcept = default;

  TemporaryFile& operator=(TemporaryFile&& other) noexcept = default;

  // `path_prefix` is the base path. The resulting filename will be this path
  // plus a unique string plus `suffix`. If `path_prefix` refers to a
  // nonexistent directory the directory will be created if possible.
  static tl::expected<TemporaryFile, std::string>
  create(const std::filesystem::path& path_prefix,
         std::string_view suffix = ".tmp");

  static bool is_tmp_file(const std::filesystem::path& path);

  // The resulting open file descriptor in read/write mode.
  Fd fd;

  // The actual filename.
  std::filesystem::path path;

private:
  TemporaryFile(Fd&& fd_, const std::filesystem::path& path_);
};

} // namespace util
